#import "MYDocument.h"


@implementation MYDocument
/*" This class encapsulates documents in a multi-document application. "*/


/*" Constant local strings "*/
#define MYDEFAULT_DOC_PATH NSLocalizedString(@"~/Untitled", @"")
#define MYCANCEL NSLocalizedString(@"Cancel", @"")
#define MYSAVE_CHANGES NSLocalizedString(\
    @"%@ has been edited. Do you want to save changes?", @"")
#define MYDEFAULT_CLOSE_ACTION NSLocalizedString(@"Close", @"")
#define MYSAVE NSLocalizedString(@"Save", @"")
#define MYDONT_SAVE NSLocalizedString(@"Don't Save", @"")
#define MYFILE_SAVE_ERROR NSLocalizedString(@"Document Save Error", @"")
#define MYFILE_SAVE_ERROR_MSG NSLocalizedString(@"Unable to save <%@>.", @"")
#define MYFILE_OPEN_ERROR NSLocalizedString(@"Document Open Error", @"")
#define MYFILE_OPEN_ERROR_MSG NSLocalizedString(@"Unable to open <%@>.", @"")


+ (NSArray *)documentExtensions
/*" Returns the extensions supported by NSImage "*/
{
  return [NSImage imageFileTypes];
}


- (void)_mySetDocumentPath:(NSString *)aPath
/*" Set the document's path by expanding tilde characters in aPath and standardizing aPath. Also sets the title of the document window. "*/
{
  NSString     *standardPath = [aPath stringByExpandingTildeInPath];

  standardPath = [standardPath stringByStandardizingPath];
  [standardPath retain];
  [_myDocumentPath release];
  _myDocumentPath = standardPath;
  [[self documentWindow] setTitleWithRepresentedFilename:_myDocumentPath];
}


- (void)_myLoadDocumentWithPath:(NSString *)aPath
  /*" Sets documents image to the image stored aPath. "*/
{
  NSImage     *imageData;

  NSAssert(![self hasEverBeenSaved], @"Attempt to load document that is already loaded.");

  if([[NSFileManager defaultManager] fileExistsAtPath:aPath])
  {
    imageData = [[NSImage alloc] initWithContentsOfFile:aPath];
  
    if(nil == imageData)
    {
      // Unable to load image data
      NSRunAlertPanel(MYFILE_OPEN_ERROR, MYFILE_OPEN_ERROR_MSG,
                      nil, nil, nil, aPath);
    }
    else
    {
      [imageView setImage:imageData];
      [imageData release];   // imageData retained by imageView
      
      // Make the image view a reasonambe size for the image displayed
      [imageView setFrameSize:[[imageView image] size]];

      // Set the path and remeber that documnet has been saved
      _myHasEverBeenSaved = YES;
      [self _mySetDocumentPath:aPath];
    }
  }
  else
  {
    // The specified path does not exist.
    // Set the document window's title but don't set its path
    [[self documentWindow] setTitleWithRepresentedFilename:aPath];
  }

}


- (id)initWithPath:(NSString *)aPath
/*" Designated Initializer: Loads the objects that represent the document from a nib file, configures the objects, and then loads the data at aPath.  Sets the document's path to a standardized version of aPath. If aPath is nil, an untitled document is created. "*/
{
  self = [super init];

  if(self)
  {
    // Load the user interface objects defined in a nib file using self as the
    // File's owner of the nib
    [NSBundle loadNibNamed:@"ImageViewerDocument.nib" owner:self];

    // Configure the loaded objects
    [imageView setTarget:self];
    [imageView setAction:@selector(noteImageWasDropped:)];
    [imageView setAutoresizingMask:NSViewNotSizable];
    [[imageView window] setReleasedWhenClosed:YES];
  
    if(nil != aPath)
    {
      [self _myLoadDocumentWithPath:aPath];
    }
    else
    {
      [self _mySetDocumentPath:MYDEFAULT_DOC_PATH];
    }
  }

  return self;
}


- (id)init
/*" Alternate Initilaizer: calls [self initWithPath:nil]. "*/
{
  return [self initWithPath:nil];
}


/*" Document management methods "*/
- (NSString *)documentPath
/*" Returns the document's path or nil. "*/
{
  return _myDocumentPath;
}


- (bool)_myShouldClose
/*" Gives users a chance to save edited documents before closing the document. Returns NO if the user cancels the close.  Returns YES otherwise. "*/
{
  BOOL     result = YES;

  if([self isDocumentEdited])
  {
    int         userChoice;
    NSString    *documentName = [self documentPath];

    if(nil == documentName)
    {
      // document has no path so use its window's title instead
      documentName = [[self documentWindow] title];
    }
    
    // the document has been edited (perhapse by a filed that was just forcibly validated
    userChoice = NSRunAlertPanel(MYDEFAULT_CLOSE_ACTION,
        MYSAVE_CHANGES, MYSAVE, MYDONT_SAVE, MYCANCEL, documentName);

    switch(userChoice)
    {
      case NSAlertDefaultReturn:
      {
        // User chooses to save changes
        [self saveDocument:nil];
        result = YES;
        break;
      }
      case NSAlertAlternateReturn:
      {
        // User chooses NOT to save changes
        result = YES;
        break;
      }
      case NSAlertOtherReturn:
      {
        // User chooses to CANCEL close
        result = NO;
        break;
      }
    }
  }

  return result;
}


- (BOOL)safeClose
/*" Gives users a chance to save edited documents before closing the document. Returns NO if the user cancels the close.  Returns YES otherwise. "*/
{
  BOOL     result = [self _myShouldClose];

  if(result)
  {
    // the user agrees to close the window
    // close the document window immediately without
    // letting the window call its -windowShouldClose: delegate 
    // method.
    [[self documentWindow] close];
  }

  return result;
}


- (void)print
/*" Shows a standard print panel and prints the document's image view. "*/
{ 
  [imageView print:self];
}


- (id)documentWindow
/*" Returns document's window "*/
{
  return [imageView window];
}

  /*" Document status "*/
- (BOOL)hasEverBeenSaved
/*" Returns YES iff document has ever been saved and has a valid path. "*/
{
  return _myHasEverBeenSaved;  
}


- (BOOL)isDocumentEdited
/*" Returns YES iff document has been edited since it was last saved. "*/
{
  return [[self documentWindow] isDocumentEdited];     
}


/*" Actions "*/
- (void)_mySaveDocumentData
/*" Save the document's image to the document's path. Displays an alert panel if the image is not saved correctly. "*/
{
  NSData     *imageData = [[imageView image] TIFFRepresentation];

  if(![imageData writeToFile:[self documentPath] atomically:YES])
  {
    // The image was not saved correctly
    NSRunAlertPanel(MYFILE_SAVE_ERROR, MYFILE_SAVE_ERROR_MSG,
                    nil, nil, nil, [self documentPath]);
  }
  else
  {
    // The image was saved correctly
    [[self documentWindow] setDocumentEdited:NO];
    _myHasEverBeenSaved = YES;
  }
}


- (IBAction)saveDocument:(id)sender
/*" Save the image data displayed by the document "*/
{
  BOOL       pathExists;
  BOOL       pathIsDirectory;
  
  // check to see if a directory exists at document path
  pathExists = [[NSFileManager defaultManager] fileExistsAtPath:[self documentPath]
      isDirectory:&pathIsDirectory];

  if(![self hasEverBeenSaved] || pathIsDirectory || nil == [self documentPath])
  {
    // The image has never been saved or it has no path or there is a directory at its path
    // Use -saveDocumentAs: because is shows a Save panel to find out where the user wants to
    // save the image
    [self saveDocumentAs:sender];
  }
  else
  {
    // Save the image data
    [self _mySaveDocumentData];
  }
  
}


- (IBAction)saveDocumentAs:(id)sender
/*" Displays a Save panel to find out where the user wants to save the document's image data. If the user does not cancel the save, this method saves the image data at the specified location. "*/
{
  NSSavePanel     *savePanel;
  NSString        *documentDirectory = [[self documentPath] stringByDeletingLastPathComponent];
  NSString        *documentName = [[self documentPath] lastPathComponent];

  savePanel = [NSSavePanel savePanel];
  [savePanel setRequiredFileType:@"tiff"];
  [savePanel setTreatsFilePackagesAsDirectories:NO];

  if (NSOKButton == [savePanel runModalForDirectory:documentDirectory file:documentName])
  {
    [self _mySetDocumentPath:[savePanel filename]];
    [self _mySaveDocumentData];
  }
}


- (IBAction)noteImageWasDropped:(id)sender
/*" Called by document's image view when the image changes. "*/
{
  // The image has changed so note that the document has been edited.
  [[self documentWindow] setDocumentEdited:YES];

  // Make the image view a reasonambe size for the image displayed
  [imageView setFrameSize:[[imageView image] size]];
}


  /*" Menu validation "*/
- (BOOL)validateMenuItem:(NSMenuItem *)aMenuItem
/*" Validate menu items that send the saveDocument: and saveDocumentAs: actions. "*/
{
  SEL 		action = [aMenuItem action];
  BOOL		enableMenuItem = NO;

  if (action == @selector(saveDocument:))
  {
    // If the document has never been saved or has been edited since it was saved
    // then enable the menu item that sends saveDocument:
    enableMenuItem = [self isDocumentEdited] || ![self hasEverBeenSaved];
  }
  else if (action == @selector(saveDocumentAs:))
  {
    // Always enable the menu item that sends saveDocumentAs:.
    enableMenuItem = YES;
  }

  return enableMenuItem;
}

/*" Window delegate methods "*/
- (BOOL)windowShouldClose:(id)sender
/*" Give users a chance to save edited documents before they close and/or cancel the close. "*/
{
  return [self _myShouldClose];
}


- (void)windowWillClose:(NSNotification *)notification
  /*" Tell document manager that document is closing. "*/
{
  imageView = nil;  // don't keep reference to closed window
  [NSApp sendAction:@selector(documentWillClose:) to:nil from:self];
}


- (void)windowDidBecomeKey:(NSNotification *)notification
/*" Tell document manager to make the receiver the active document "*/
{
  [NSApp sendAction:@selector(makeDocumentActive:) to:nil from:self];
}


- (void)dealloc
/*" Clean-up "*/
{
  [_myDocumentPath release];
  [super dealloc];
}

@end
