#import "MYDocumentManager.h"
#import "MYDocument.h"


@implementation MYDocumentManager
/*" A single instance of this class is used as the application's delegate and manages documents in a multi-document application.

There is a single "active" document at any time.  The "active" document is usually the document that controls that last key window, but the active document can be set programatically.
"*/


/*" Constant local strings "*/
#define MYQUIT NSLocalizedString(@"Quit", @"")
#define MYOPEN NSLocalizedString(@"Open Error", @"")
#define MYOPEN_DOCUMENT_ERROR_MSG NSLocalizedString(@"Unable to open document.", @"")
#define MYNEW_DOCUMENT_ERROR_MSG NSLocalizedString(@"Unable to create new document.", @"")
#define MYUNSAVED_DOCS_MSG NSLocalizedString(@"There are unsaved documents.\nReview them?", @"")
#define MYCANCEL NSLocalizedString(@"Cancel", @"")
#define MYQUIT_ANYWAY NSLocalizedString(@"Quit Anyway", @"")
#define MYREVIEW_UNSAVED NSLocalizedString(@"Review", @"")
#define MYUNTITLED NSLocalizedString(@"Untitled", @"The name of untitled documents")

- (void)_myInitInstanceVariables
/*" Initialize instance variables: called by -init and -awakeFromNib. "*/
{
  NSRect    screenVisibleFrame = [[NSScreen mainScreen] visibleFrame];
  
  // Create array of open documents
  _myOpenDocuments = [[NSMutableArray allocWithZone:[self zone]] init];
  
  // Set initial cascade point
  _myWindowCascadePoint = NSMakePoint(0, NSMaxY(screenVisibleFrame));
}


- (id)init
/*" Designated initializer "*/
{
  self = [super init];

  if(nil != self)
  {
    [self _myInitInstanceVariables];
  }

  return self;
}


- (void)awakeFromNib
/*" Called automatically after receiver is fully loaded from a nib file. "*/
{
  [self _myInitInstanceVariables];
}


- (Class)documentClass
/*" Returns the class used to encapsulate documents "*/
{
  return [MYDocument class];
}


- (NSArray *)documentExtensions;
/*" Returns an array of NSString file extensions for documents that can be opened. This implementation returns an array containing elements obtained by calling [[self documentClass] documentExtensions].  Override this method to provide more sophisticated behavior and support multiple document classes. To allow any opening of documents with any extension or no extension, override this method to return an empty array. "*/
{
  NSMutableArray     *supportedExtensions = [NSMutableArray array];
  Class              documentClass = [self documentClass];

  if([documentClass respondsToSelector:@selector(documentExtensions)])
  {
    [supportedExtensions addObjectsFromArray:[documentClass documentExtensions]];
  }

  return supportedExtensions;
}

  /*" Manipulate documents "*/
- (void)makeDocumentsPerformSelector:(SEL)aSelector
/*" Sends the aSelector message to all open documents. "*/
{
  [[self existingOpenDocuments] makeObjectsPerformSelector:aSelector];
}


- (void)makeDocumentsPerformSelector:(SEL)aSelector withObject:(id)anArg
/*" Sends the aSelector message to all open documents using anArg as an argument to the message. "*/
{
  [[self existingOpenDocuments] makeObjectsPerformSelector:aSelector withObject:anArg];
}


- (NSArray *)existingOpenDocuments
/*" Returns an array of open documets. "*/
{
  return _myOpenDocuments;
}


- (id)existingOpenDocumentWithPath:(NSString *)aPath
/*" Returns the open document with the specified path or nil. "*/
{
  NSString 		*standardPath = [aPath stringByStandardizingPath];
  id			result = nil;
  id			currentDocument = nil;
  NSEnumerator  *enumerator = [_myOpenDocuments objectEnumerator];

  while(nil == result && nil != (currentDocument = [enumerator nextObject]))
  {
    if([currentDocument respondsToSelector:@selector(documentPath)])
    {
      // the current document has a path
      if([[currentDocument documentPath] isEqualToString:standardPath])
      {
        // currentDocument has aPath
        result = currentDocument;
      }
    }
  }

  return result;
}


- (id)activeDocument
/*" Returns the active document or nil.  The active document is usually the last document who's window was key. "*/
{
  // The last object in _myOpenDocuments is assumed to be the active document
  return (0 < [_myOpenDocuments count]) ? [_myOpenDocuments lastObject] : nil;
}


- (NSString *)_myNextUntitledDocumentName
  /*" Returns an NSString with a unique untitled document name by appending an integer to the string "Untitled" and the first document extension in the array returned by -documentExtensions. "*/
{
  static int    lastUntitledIndex = 1;
  NSString      *result = nil;
  NSString      *extension = @"";
  NSArray       *documentExtensions = [self documentExtensions];

  if(0 < [documentExtensions count])
  {
    // documentExtensions has at least 1 element so use first element as
    // default untitled document extension
    extension = [documentExtensions lastObject];
  }

  do {
    [result release];		// release string from previous iteration (nil on first iteration)

    result = [[NSString alloc] initWithFormat:@"~/%@%d.%@", MYUNTITLED,
      lastUntitledIndex++, extension];

  } while([[NSFileManager defaultManager] fileExistsAtPath:[result stringByStandardizingPath]]);

  [result autorelease];

  return result;
}


- (void)_myRegisterDocument:(id)aDocument
/*" Adds aDocument to the receiver's array of open documents, makes aDocument's window visible on screen, and makes aDocument the active document. "*/
{
  [_myOpenDocuments addObject:aDocument];

  if([aDocument respondsToSelector:@selector(documentWindow)])
  {
    id     documentWindow = [aDocument documentWindow];
    
    // Cascade the registered document's window
    _myWindowCascadePoint = [documentWindow cascadeTopLeftFromPoint:_myWindowCascadePoint];

    // make registered document's window visible and key
    [documentWindow makeKeyAndOrderFront:self];
  }
}


- (void)_myOpenDocumentWithPath:(NSString *)aPath
{
  id            newDocument = [self existingOpenDocumentWithPath:aPath];

  if(nil != newDocument)
  { // an existing open document already has currentPath
    [[newDocument documentWindow] makeKeyAndOrderFront:self];
  }
  else
  { // no open document already has currentPath
    newDocument = [[[self documentClass] allocWithZone:[self zone]]
              initWithPath:aPath];
    if(nil == newDocument)
    {
      NSLog(@"%@", MYOPEN_DOCUMENT_ERROR_MSG);
    }
    else
    {
      [self _myRegisterDocument:newDocument];  // Retains newDocument
      [newDocument release];
    }
  }
}


- (IBAction)openDocument:(id)sender
/*" Displays an open panel that alows selections of multiple paths and opens the selected files. In this implementation, individual files can be selected and directories can not. The paths that are selected must have one of the extensions in the array returned by -documentExtensions. If the array returned from -documentExtensions is empty, any path is accepted. "*/
{
  NSArray       *typesArray;
  NSOpenPanel   *openPanel;
  NSString      *directory;
  id            activeDocument;

  typesArray = [self documentExtensions];
  if(0 == [typesArray count])
  {
    // Set fileTypesArray to nil so that any type is accepted
    typesArray = nil;
  }
  activeDocument = [self activeDocument];

  if(nil != activeDocument && [activeDocument respondsToSelector:@selector(documentPath)])
  {
    // Use active document's directory as initial directory
    directory = [[activeDocument documentPath] stringByDeletingLastPathComponent];
  } else {
    directory = @"";
  }
  
  // Configure the open panel
  openPanel = [NSOpenPanel openPanel];
  [openPanel setAllowsMultipleSelection:YES];
  [openPanel setTreatsFilePackagesAsDirectories:NO];
  [openPanel setDirectory:directory];
  [openPanel setCanChooseDirectories:NO];
  [openPanel setCanChooseFiles:YES];
  [openPanel setResolvesAliases:YES];

  // Catch any exceptions raised while opening a document
  if (NSOKButton == [openPanel runModalForTypes:typesArray])
  {
    NSArray       *selectedPaths = [openPanel filenames];
    NSEnumerator  *enumerator = [selectedPaths objectEnumerator];
    NSString      *currentPath;

    while(nil != (currentPath = [enumerator nextObject]))
    {
      [self _myOpenDocumentWithPath:currentPath];
    }
  }
}


- (IBAction)newDocument:(id)sender
  /*" Creates a new instance of the class returned from the -documentClass.  The new document is given a unique title and made the active document. "*/
{
  id			newDocument;
  NSString      *newDocumentPath = [self _myNextUntitledDocumentName];
  
  newDocument = [[[self documentClass] allocWithZone:[self zone]] initWithPath:newDocumentPath];
  if(nil == newDocument)
  {
    NSLog(@"%@", MYNEW_DOCUMENT_ERROR_MSG);
  }
  else
  {
    [self _myRegisterDocument:newDocument];  // Retains newDocument
    [newDocument release];
  }
}


- (IBAction)saveAllDocuments:(id)sender
/*" This method sends the -saveDocument: message to all open documents. "*/
{
  [self makeDocumentsPerformSelector:@selector(saveDocument:) withObject:sender];
}


/*" Document communication actions "*/
- (void)documentWillClose:(id)sender
  /*" Removes sender from receiver's open documents array "*/
{
  [_myOpenDocuments removeObject:sender];
}


- (void)makeDocumentActive:(id)sender
/*" Sets the receiver's active document to sender if possible.  If sender is not an open document, this method does nothing. This method does not make sender's window key. "*/
{

  if(0 < [_myOpenDocuments count])
  {
    // There is at least 1 open document
    int       index = [_myOpenDocuments indexOfObject:sender];

    if(index != NSNotFound)
    {
      // anObject is an open document
      // Swap positions in _myOpenDocuments: the last object in _myOpenDocuments is
      // assumed to be the active document
      [sender retain];
      [_myOpenDocuments replaceObjectAtIndex:index withObject:[_myOpenDocuments lastObject]];
      [_myOpenDocuments removeLastObject];
      [_myOpenDocuments addObject:sender];
      [sender release];
    }
  }
}


/*" Menu validation "*/
- (BOOL)validateMenuItem:(NSMenuItem *)aMenuItem
/*" Validate menu items that send the openDocument:, newDocument:, or saveAllDocuments: actions. "*/
{
  SEL 		action;
  BOOL		enableMenuItem = NO;

  action = [aMenuItem action];
  
  // The -open:, -new: and -saveAll: actions are provided by this calss and
  // therefore validated by this class.
  if (action == @selector(openDocument:) || action == @selector(newDocument:))
  {
    enableMenuItem = YES;
  }
  else if (action == @selector(saveAllDocuments:))
  {
    enableMenuItem = (0 < [_myOpenDocuments count]);
  }

  return enableMenuItem;
}


/*" Application delegate methods "*/
- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
/*" Implemented to Give user a chance to review and save any unsaved documents before terminating "*/
{
  NSEnumerator                *enumerator;
  id                          currentDocument;
  int                         choice;
  NSApplicationTerminateReply result = NSTerminateNow;
  BOOL                        foundUnsaved = NO;

  NS_DURING
    // Catch exception during review and save
    
    // Determine if theer are any unsaved documents
    enumerator = [_myOpenDocuments objectEnumerator];
    while(!foundUnsaved && nil != (currentDocument = [enumerator nextObject]))
    {
      if([currentDocument respondsToSelector:@selector(isDocumentEdited)])
      {
        // Found at least one unsaved document
        foundUnsaved = [currentDocument isDocumentEdited];
      }
    }

    if(foundUnsaved)
    {
      // Find out of the user wants to review and possibly save the unsaved documents,
      // cancel the termination, or terminate anyway
      choice = NSRunAlertPanel(MYQUIT, MYUNSAVED_DOCS_MSG, MYREVIEW_UNSAVED,
                               MYQUIT_ANYWAY, MYCANCEL);
      if (choice == NSAlertOtherReturn)
      { // User selected Cancel
        result = NSTerminateCancel;
      }
      else if (choice != NSAlertAlternateReturn)
      {	// User selected Review Unsaved
        // Give the user the chance to review the edited document(s). */
        enumerator = [_myOpenDocuments objectEnumerator];
        while(result == NSTerminateNow && nil != (currentDocument = [enumerator nextObject]))
        {
          if([currentDocument respondsToSelector:@selector(safeClose)])
          {
            // Cause unsaved documents to show a save panel
            if(![currentDocument safeClose])
            { // User selected Cancel
              result = NSTerminateCancel;
            }
          }
        }
      }
      else
      { // User chooses to quit withot saving
      }
    }
    NS_HANDLER
      NSLog(@"%@: %@", [localException name], [localException reason]);
    NS_ENDHANDLER

    return result;
}


- (BOOL)application:(NSApplication *)sender openFile:(NSString *)filename
/*" Called automatically when application is requested to open a document either by drag and drop or double clicking in finder. Returns YES if the file is successfully opened, and NO otherwise "*/
{
  int				result = NO;

  if([[self documentExtensions] containsObject:[filename pathExtension]])
  {
    // Filename has an accepted extension
    [self _myOpenDocumentWithPath:filename];
    result = YES;
  }

  return result;
}


- (BOOL)application:(NSApplication *)sender openTempFile:(NSString *)filename
/*" Called automatically when application is requested to open a temporary document. This method is implemented to call [self application:app openFile:filename]. "*/
{
  return [self application:sender openFile:filename];
}


- (BOOL)applicationShouldOpenUntitledFile:(NSApplication *)sender
/*" Called automatically: returns YES meaning that it is OK to open an untitled document. "*/
{
  return YES;
}


- (BOOL)applicationOpenUntitledFile:(NSApplication *)sender
/*" Called automatically when application is requested to open an untitled document. This method is implemented to call [self newDocument:nil] "*/
{
  [self newDocument:nil];

  return YES;
}


- (BOOL)application:(NSApplication *)sender printFile:(NSString *)filename
/*" Opens the document at filename if possible and prints it to the user's default printer.  If the user does not have a default printer, a print panel is raised. "*/
{
  BOOL      result;

  result = [self application:sender openFile:filename];
  if(result)
  {
    id        document;

    document = [self existingOpenDocumentWithPath:filename];
    if([document respondsToSelector:@selector(print)])
    {
      [document print];
    }
  }

  return result;
}


- (void)dealloc
/*" Claen-up "*/
{
  [_myOpenDocuments release];
  _myOpenDocuments = nil;

  [super dealloc];
}

@end
